"use strict";
const hrtime_now_1 = require("hrtime-now");
const typescript_ast_util_1 = require("typescript-ast-util");
const typescript_plugin_util_1 = require("typescript-plugin-util");
const ts_simple_ast_1 = require("ts-simple-ast");
const methodDelegate_1 = require("./methodDelegate");
const PLUGIN_NAME = 'typescript-plugin-method-delegate';
const REFACTOR_ACTION_NAME = `${PLUGIN_NAME}-refactor-action`;
let ts;
let info;
const pluginDefinition = { getApplicableRefactors, getEditsForRefactor };
let target;
function getApplicableRefactors(fileName, positionOrRange, userPreferences) {
    const t0 = hrtime_now_1.now();
    const refactors = info.languageService.getApplicableRefactors(fileName, positionOrRange, userPreferences) || [];
    const program = info.languageService.getProgram();
    const sourceFile = program.getSourceFile(fileName);
    if (!sourceFile) {
        return refactors;
    }
    target = typescript_ast_util_1.findChildContainingRange(sourceFile, typescript_ast_util_1.positionOrRangeToRange(positionOrRange));
    if (!(target && [ts.SyntaxKind.PropertySignature, ts.SyntaxKind.PropertyDeclaration].includes(target.kind))) {
        info.project.projectService.logger.info(`${PLUGIN_NAME} no getEditsForRefactor because findChildContainedRange undefined, target.kind == ${typescript_ast_util_1.getKindName(target.kind) + ' - ' + [ts.SyntaxKind.PropertySignature, ts.SyntaxKind.PropertyDeclaration].join(', ')}`);
        return refactors;
    }
    if (![ts.SyntaxKind.PropertySignature, ts.SyntaxKind.PropertyDeclaration].includes(target.kind)) {
        target = target.parent;
    }
    const name = target.name && target.name.getText() || '';
    refactors.push({
        name: `${PLUGIN_NAME}-refactor-info`,
        description: 'Add type',
        actions: [
            { name: REFACTOR_ACTION_NAME, description: 'Delegate methods to property"' + name + '"' }
        ],
    });
    info.project.projectService.logger.info(`${PLUGIN_NAME} getApplicableRefactors took ${hrtime_now_1.now() - t0}`);
    return refactors;
}
function getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, userPreferences) {
    const t0 = hrtime_now_1.now();
    const refactors = info.languageService.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, userPreferences);
    if (actionName != REFACTOR_ACTION_NAME || !target) {
        return refactors;
    }
    try {
        const project = typescript_plugin_util_1.createSimpleASTProject(info.project);
        const descAtPo = project.getSourceFileOrThrow(fileName).getDescendantAtPos(typescript_ast_util_1.positionOrRangeToNumber(positionOrRange));
        const typeDeclaration = descAtPo.getParent().getParent();
        const propertyDeclaration = descAtPo.getParent();
        if (ts_simple_ast_1.TypeGuards.isInterfaceDeclaration(typeDeclaration) && ts_simple_ast_1.TypeGuards.isPropertySignature(propertyDeclaration)) {
            methodDelegate_1.methodDelegateOnInterface(typeDeclaration, propertyDeclaration);
            project.emit();
            project.saveSync();
        }
        else if (ts_simple_ast_1.TypeGuards.isClassDeclaration(typeDeclaration) && ts_simple_ast_1.TypeGuards.isPropertyDeclaration(propertyDeclaration)) {
            methodDelegate_1.methodDelegateOnClass(typeDeclaration, propertyDeclaration);
            project.emit();
            project.saveSync();
        }
        else {
            info.project.projectService.logger.info(`${PLUGIN_NAME} getEditsForRefactor not executed because typeguards didn't applied for typeDeclaration:${typeDeclaration.getKindName()}, propertyDeclaration:${propertyDeclaration.getKindName()}`);
        }
        info.project.projectService.logger.info(`${PLUGIN_NAME} getEditsForRefactor took ${hrtime_now_1.now() - t0}`);
    }
    catch (error) {
        info.project.projectService.logger.info(`${PLUGIN_NAME} getEditsForRefactor error: ${error.toString()} - stack: ${error.stack}`);
    }
    return refactors;
}
module.exports = typescript_plugin_util_1.getPluginCreate(pluginDefinition, (modules, anInfo) => {
    ts = modules.typescript;
    info = anInfo;
    info.project.projectService.logger.info(`${PLUGIN_NAME} created`);
});
//# sourceMappingURL=index.js.map